#include "mandelbrot.h"
#include "progressBar.h"
#include <sstream>

auto Mandelbrot::hsv2rgb( dType h, dType s, dType v ) const -> Color {
    h -= std::floor( h );
    h *= 6.0;
    if ( s < 0 ) { s = 0; }
    if ( s > 1 ) { s = 1; }
    if ( v < 0 ) { v = 0; }
    if ( v > 1 ) { v = 1; }
    dType f = h - std::floor( h );
    Color c;
    switch ( int( std::floor( h ) ) ) {
    case 0:
        c = Color( 1, 1 - ( 1 - f )*s, 1 - s );
        break;
    case 1:
        c = Color( 1 - s * f, 1, 1 - s );
        break;
    case 2:
        c = Color( 1 - s, 1, 1 - ( 1 - f ) * s );
        break;
    case 3:
        c = Color( 1 - s, 1 - s * f, 1 );
        break;
    case 4:
        c = Color( 1 - ( 1 - f ) * s, 1 - s, 1 );
        break;
    default:
        c = Color( 1, 1 - s, 1 - s * f );
    }
    c.r *= v;
    c.g *= v;
    c.b *= v;
    return c;
}

auto Mandelbrot::mappedPoint( Point p ) const -> Point {
    return{
        ( p.X - _area.X / 2 ) * _zoom + _center.X,
        ( _area.Y / 2 - p.Y ) * _zoom + _center.Y
    };
}
auto Mandelbrot::palette( dType d ) const -> Color {
    dType v = std::sin( d * 0.1847969 ) / 2 + 0.5;
    return hsv2rgb(
        d * 0.00521336,
        ( std::sin( d * 0.162012467 ) / 2 + 0.5 ) * ( 1 - v ),
        v
        );
}

auto Mandelbrot::interpolate( Color a, Color b, dType step ) const -> Color {
    return{
        a.r + step * ( b.r - a.r ),
        a.g + step * ( b.g - a.g ),
        a.b + step * ( b.b - a.b )
    };
}

auto Mandelbrot::logColor( dType d ) const -> Color {
    if ( d < 0 )
        return Color();

    d = std::log( d + 50 ) * 100;
    dType p = std::floor( d );
    return interpolate( palette( p ), palette( p + 1 ), d - p );
}

auto Mandelbrot::computePoint( Point p ) const -> dType {
    Complex z( p );
    Complex c( p );

    unsigned i = 0;
    for ( ; i < _maxIter; ++i ) {
        z = z*z + c;
        if ( std::abs( z ) >= 2 )
            break;
    }

    if ( std::abs( z ) >= 2 ) {
        z = z*z + c;
        z = z*z + c;
        return i + 10.0 - std::log2( std::log( std::abs( z ) ) );
    }
    return -1.0;
}

auto Mandelbrot::colorizePoint( PrecisePoint p ) const -> Color {
    return logColor( computePoint( mappedPoint( p ) ) );
}

auto Mandelbrot::draw() -> void {
    ProgressBar bar( std::cout, _name, _area.Y );
    for ( size_t y = 0; y < _area.Y; ++y ) {
        for ( size_t x = 0; x < _area.X; ++x )
            _picture << colorizePoint( PrecisePoint( x, y ) ).normalize();
        ++bar;
    }
}

auto Mandelbrot::parseName( const std::string &name ) const -> std::string {
    std::ostringstream oss;

    bool special = false;
    for ( char c : name ) {
        if ( special ) {
            special = false;
            switch ( c ) {
            case 'x':
                oss << _center.X;
                break;
            case 'y':
                oss << _center.Y;
                break;
            case 'r':
                oss << ( std::min( _area.X, _area.Y ) * _zoom ) / 2;
                break;
            case 'a':
                oss << _area.X << 'x' << _area.Y;
                break;
            case 'p':
                oss << _center.X << ';' << _center.Y;
                break;
            case 'i':
                oss << _maxIter;
                break;
            default:
                oss << c;
                break;
            }
        }
        else if ( c == '$' )
            special = true;
        else
            oss << c;
    }
    return oss.str();
}
