#pragma once

#include <complex>
#include <algorithm>
#include "bmpWriter.h"

class Mandelbrot {
public:
    using dType = long double;
    using Complex = std::complex< dType >;

    struct Color {
        dType r;
        dType g;
        dType b;

        Color( dType r, dType g, dType b ) :
            r( r ), g( g ), b( b ) {}
        Color() :
            r( 0 ), g( 0 ), b( 0 ) {}

        BitmapWriter::RGB normalize() const {
            using Byte = BitmapWriter::RGB::Byte;
            return{ Byte( r * 255 ), Byte( g * 255 ), Byte( b * 255 ) };
        }
    };
    struct PrecisePoint {
        unsigned X;
        unsigned Y;

        PrecisePoint( unsigned x, unsigned y ) :
            X( x ), Y( y ) {}
        PrecisePoint( const PrecisePoint & ) = default;
        PrecisePoint &operator=( const PrecisePoint & ) = default;

        operator Complex() const {
            return Complex( X, Y );
        }
    };
    struct Point {
        dType X;
        dType Y;

        Point( dType x, dType y ) :
            X( x ), Y( y ) {}

        Point( const PrecisePoint &pp ) :
            X( pp.X ), Y( pp.Y ) {}

        Point( const Point & ) = default;
        Point &operator=( const Point & ) = default;

        Point( const Complex &c ) :
            X( c.real() ), Y( c.imag() ) {}

        operator Complex() const {
            return{ X, Y };
        }

        PrecisePoint floor() const {
            return{
                unsigned( std::floor( X ) ),
                unsigned( std::floor( Y ) )
            };
        }
    };

    struct Settings {
        PrecisePoint area;
        Point center;
        dType radius;
        unsigned maxIterations;
        std::string name;

        dType zoom() const {
            return 2 * radius / ( std::min( area.X, area.Y ) );
        }

        Settings( PrecisePoint area = { 1600, 900 },
                  Point center = { 0, 0 },
                  dType radius = 1,
                  unsigned maxIterations = 1000,
                  std::string name = "mandelbrot-$x;$y;$r[$a].bmp" )
                  :
                  area( area ),
                  center( center ),
                  radius( radius ),
                  maxIterations( maxIterations ),
                  name( name ) {}

    };

    Mandelbrot( const Settings &s ) :
        _area( s.area ),
        _center( s.center ),
        _zoom( s.zoom() ),
        _maxIter( s.maxIterations )
    {
        _name = parseName( s.name );
        _picture.open( _name, _area.X, _area.Y );
    }
    void draw();

private:
    PrecisePoint _area;
    Point _center;
    dType _zoom;
    unsigned _maxIter;
    std::string _name;
    BitmapWriter _picture;

    Color hsv2rgb( dType h, dType s, dType v ) const;
    Point mappedPoint( Point p ) const;
    Color palette( dType d ) const;
    Color interpolate( Color a, Color b, dType step ) const;
    Color logColor( dType d ) const;
    dType computePoint( Point p ) const;
    Color colorizePoint( PrecisePoint p ) const;
    std::string parseName( const std::string &name ) const;
};

static inline std::istream &operator>>( std::istream &in, Mandelbrot::PrecisePoint &p ) {
    in >> p.X >> p.Y;
    return in;
}
static inline std::istream &operator>>( std::istream &in, Mandelbrot::Point &p ) {
    in >> p.X >> p.Y;
    return in;
}

static inline std::istream &operator>>( std::istream &in, Mandelbrot::Settings &s ) {
    in >> s.area >>
        s.center >>
        s.radius >>
        s.maxIterations >>
        s.name;
    return in;
}

